from Extension import Extension


class DocumentObjectExtension(Extension):
    """
    Base class for all document object extensions
    Author: Stefan Troeger (stefantroeger@gmx.net)
    Licence: LGPL
    """
    ...
