# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 Furgo                                              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import Arch
import Draft
import FreeCAD as App
from bimtests import TestArchBase


class TestArchFrame(TestArchBase.TestArchBase):

    def test_makeFrame(self):
        """Test the makeFrame function."""

        obj = Arch.makeFrame(None, None)
        self.assertIsNotNone(obj, "makeFrame failed to create an object")
        self.assertEqual(obj.Label, "Frame", "Incorrect default label for Frame")

    def testFrame(self):
        l = Draft.makeLine(App.Vector(0, 0, 0), App.Vector(-2, 0, 0))
        p = Draft.makeRectangle(length=0.5, height=0.5)
        f = Arch.makeFrame(l, p)
        self.assertTrue(f, "Arch Frame failed")
