/***************************************************************************
 *   Copyright (c) 2016 WandererFan <wandererfan@gmail.com>                *
 *                                                                         *
 *   This file is part of the FreeCAD CAx development system.              *
 *                                                                         *
 *   This library is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU Library General Public           *
 *   License as published by the Free Software Foundation; either          *
 *   version 2 of the License, or (at your option) any later version.      *
 *                                                                         *
 *   This library  is distributed in the hope that it will be useful,      *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Library General Public License for more details.                  *
 *                                                                         *
 *   You should have received a copy of the GNU Library General Public     *
 *   License along with this library; see the file COPYING.LIB. If not,    *
 *   write to the Free Software Foundation, Inc., 59 Temple Place,         *
 *   Suite 330, Boston, MA  02111-1307, USA                                *
 *                                                                         *
 ***************************************************************************/

#ifndef DRAWINGGUI_QGRAPHICSITEMVIEWSPREADSHEET_H
#define DRAWINGGUI_QGRAPHICSITEMVIEWSPREADSHEET_H

#include <Mod/TechDraw/TechDrawGlobal.h>

#include "QGIViewSymbol.h"
#include "QGIUserTypes.h"

namespace TechDraw {
class DrawViewSpreadsheet;
}

namespace TechDrawGui
{

class TechDrawGuiExport QGIViewSpreadsheet : public QGIViewSymbol
{
public:
    QGIViewSpreadsheet();
    ~QGIViewSpreadsheet() override = default;

    enum {Type = UserType::QGISpreadsheet};
    int type() const override { return Type;}

    void setViewFeature(TechDraw::DrawViewSpreadsheet *obj);

};

} // end namespace TechDrawGui

#endif // DRAWINGGUI_QGRAPHICSITEMVIEWSPREADSHEET_H
