from GeoFeature import GeoFeature


class Part(GeoFeature):
    """
    Author: Juergen Riegel (FreeCAD@juergen-riegel.net)
    Licence: LGPL
    This class handles document objects in Part
    """
    ...
