#!/bin/sh
dnf --assumeyes install boost-devel cmake Coin3 Coin3-devel desktop-file-utils doxygen \
    eigen3-devel freeimage-devel freetype freetype-devel gcc gettext git libspnav-devel \
    libXmu-devel med med-devel mesa-libGLU-devel ode-devel opencascade-devel \
    opencv-devel openmpi-devel pcl-devel pyside2-tools python3 python3-devel \
    python3-matplotlib python3-pivy python3-pyside2-devel qt5-qtsvg-devel \
    qt5-qttools-static qt-devel \
    smesh-devel SoQt-devel swig tbb-devel vtk xerces-c xerces-c-devel

